use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.006007
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib  ) );
__DATA__
AES
Don'ts
Fu
HMAC
Liu
MAC's
Rekeying
Sereal
al
authenticator
ciphertext
cryptographically
decrypting
et
login
scalability
unencrypted
David
Golden
dagolden
Tom
Hukins
tom
lib
Session
Storage
Secure
